@@ -1,4 +1,5 @@
 use crate::IConfiguration;
+use serde_json::json;
 use std::str::FromStr;
 use std::time::Duration;
 
@@ -33,6 +34,28 @@ pub struct Scopes {
     scopes: Vec<Scope>,
 }
 
+#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
+pub struct RedeemPayload {
+    pub code: String,
+    pub client_id: url::Url,
+    pub redirect_uri: url::Url,
+    pub code_verifier: String,
+}
+
+#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
+pub struct Profile {
+    pub name: Option<String>,
+    pub url: Option<url::Url>,
+    pub photo: Option<url::Url>,
+    pub email: Option<String>,
+}
+
+#[derive(serde::Serialize, serde::Deserialize, Debug, Clone)]
+pub struct CodeResponse {
+    me: url::Url,
+    profile: Option<Profile>,
+}
+
 struct ScopesVisitor {}
 
 impl<'de> Visitor<'de> for ScopesVisitor {
@@ -192,14 +215,15 @@ pub struct CodeVerificationRequest {
 
 #[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
 pub struct Token {
+    pub me: url::Url,
+    pub access_token: AccessToken,
+
     #[serde(rename = "scope")]
     pub scopes: Scopes,
-    pub access_token: AccessToken,
-    pub expires_in: Option<Duration>,
-    pub refresh_token: Option<RefreshToken>,
 
-    pub me: String,
-    pub profile: Option<std::collections::HashMap<String, String>>,
+    pub expires_in: Option<u64>,
+    pub refresh_token: Option<RefreshToken>,
+    pub profile: Option<Profile>,
 }
 
 impl TokenResponse<BasicTokenType> for Token {
@@ -214,7 +238,7 @@ impl TokenResponse<BasicTokenType> for Token {
         self.refresh_token.as_ref()
     }
     fn expires_in(&self) -> Option<Duration> {
-        self.expires_in
+        self.expires_in.map(|s| Duration::from_secs(s))
     }
     fn token_type(&self) -> &BasicTokenType {
         &BasicTokenType::Bearer
@@ -245,6 +269,7 @@ impl CodeVerificationRequest {
     }
 }
 
+/// Confirms that the provided token is valid with the provided endpoint.
 pub async fn verify<C>(cfg: &C, endpoint: &str, access_token: &str) -> Option<Token>
 where
     C: IConfiguration + Send + Sync,
@@ -272,6 +297,47 @@ where
     }
 }
 
+pub async fn redeem<'a, C, R>(
+    cfg: &C,
+    endpoint: &str,
+    redeem_payload: RedeemPayload,
+) -> Result<R, crate::Error>
+where
+    C: IConfiguration + Send + Sync,
+    R: for<'de> Deserialize<'de>,
+{
+    let endpoint_resp = cfg
+        .request(Some(
+            "Requesting information about an authorization request".to_owned(),
+        ))
+        .request(reqwest::Method::POST, endpoint.to_owned())
+        .header("accept", "application/json")
+        .body(
+            json!({
+                "grant_type": "authorization_code",
+                "code": redeem_payload.code,
+                "client_id": redeem_payload.client_id,
+                "redirect_uri": redeem_payload.redirect_uri,
+                "code_verifier": redeem_payload.code_verifier
+            })
+            .to_string(),
+        )
+        .send()
+        .await
+        .map_err(|e| crate::Error::Other(anyhow::anyhow!(e)))?;
+
+    let resp = endpoint_resp
+        .error_for_status()
+        .map_err(|e| anyhow::anyhow!(e))?;
+    log::info!("Handling response from endpoint.");
+    let text = resp
+        .text()
+        .await
+        .map_err(|e| crate::Error::Other(anyhow::anyhow!(e)))?;
+
+    serde_json::from_str(text.as_str()).map_err(|e| crate::Error::Other(anyhow::anyhow!(e)))
+}
+
 #[cfg(test)]
 #[allow(non_snake_case)]
 mod tests {
@@ -384,11 +450,9 @@ mod tests {
         assert!(matches!(result, Ok(_)));
 
         let token = result.unwrap();
-        assert_eq!(token.me, "https://jacky.wtf/".to_owned());
-        assert!(matches!(token.profile, Some(_)));
-
+        assert!(token.profile.is_some());
         let profile = token.profile.unwrap();
-        assert_eq!(profile.get("url"), Some(&"https://jacky.wtf/".to_owned()));
+        assert_eq!(profile.url, "https://jacky.wtf/".parse().ok());
     }
 
     #[tokio::test(flavor = "multi_thread")]
@@ -412,11 +476,51 @@ mod tests {
         assert!(result.is_some());
 
         let token = result.unwrap();
-        assert_eq!(token.me, "https://jacky.wtf/".to_owned());
+        assert_eq!(token.me, "https://jacky.wtf/".parse().unwrap());
         assert_eq!(token.access_token.secret().clone(), "a-token".to_owned());
         assert_eq!(token.scopes.to_string(), "read profile email".to_owned());
     }
 
+    #[tokio::test(flavor = "multi_thread")]
+    async fn redeem_resolves_responses() {
+        let cfg = test::init();
+        let endpoint = format!("{}/endpoint", mockito::server_url());
+        let body = serde_json::json!({
+            "me": "https://jacky.wtf/",
+            "token_type": "bearer",
+            "scope": "read profile email create",
+            "access_token": "the-access-token",
+            "profile": {
+                "name": "Jacky Alciné",
+                "url": "https://jacky.wtf/",
+                "photo": "https://jacky.wtf/photo.jpeg",
+                "email": "yo+codespam@jacky.wtf"
+            }
+
+        })
+        .to_string();
+
+        let endpoint_mock = mockito::mock("POST", "/endpoint")
+            .with_body(&body)
+            .expect_at_most(2)
+            .create();
+        let payload = RedeemPayload {
+            code: "a-code".to_owned(),
+            code_verifier: "a-code-verifier".to_owned(),
+            client_id: "https://indieweb.org".parse().unwrap(),
+            redirect_uri: "https://indieweb.org/back".parse().unwrap(),
+        };
+
+        let optional_token: Result<super::Token, crate::Error> =
+            super::redeem(&cfg, &endpoint, payload.clone()).await;
+        let optional_code: Result<super::CodeResponse, crate::Error> =
+            super::redeem(&cfg, &endpoint, payload.clone()).await;
+        endpoint_mock.assert();
+
+        assert!(optional_code.is_ok());
+        assert!(optional_token.is_ok());
+    }
+
     #[test]
     fn Scopes_assert_presence() -> Result<(), anyhow::Error> {
         assert!(!Scopes::from_str("read profile")?.has("update"));
